from src.holiday_events import EventFactory
from src.utils import load_config
from snowflake.snowpark.session import Session
import snowflake.snowpark.functions as F
import snowflake.snowpark.types as T
import polars as pl
import pandas as pd

config = load_config("config/snowpark_transformer_config.yml")

# Establish Snowpark session:
session = Session.builder.configs(config).create()

# Params:
table_name = "WALMART_CURRENT"
date_col = "DATE"

# Get target table:
df = session.table(table_name)
df = df.with_column(date_col, F.to_date(F.col(date_col)))

# Calculate date range:
date_range = df.agg(
    F.min(F.col(date_col)).alias("min_date"),
    F.max(F.col(date_col)).alias("max_date"),
).to_pandas()

dates = pd.date_range(date_range["MIN_DATE"][0], date_range["MAX_DATE"][0])

# Create holiday features:
factory = EventFactory(country="US", date_col=date_col)
events = factory.get_events(
    years=range(dates[0].year, dates[-1].year + 1),
    add_special=True,
    lower_window=3,
    upper_window=1,
)
events = factory.time_to_events(dates=dates, events=events)

# Type conversions:
events = events.with_columns(
    **{
        "DATE": pl.col("DATE").cast(str),
        "time_to_next_event": pl.col("time_to_next_event").dt.days(),
        "time_since_last_event": pl.col("time_since_last_event").dt.days(),
    }
)

# Export to Snowflake:
table_to_sql = session.create_dataframe(
    events.to_pandas(),
    schema=T.StructType(
        [
            T.StructField("DATE", T.DateType),
            T.StructField("lower_window", T.IntegerType),
            T.StructField("upper_window", T.IntegerType),
            T.StructField("next_event", T.StringType),
            T.StructField("time_to_next_event", T.IntegerType),
            T.StructField("last_event", T.StringType),
            T.StructField("time_since_last_event", T.IntegerType),
        ]
    ),
)
table_to_sql.show()
table_to_sql.write.save_as_table("HOLIDAY_FEATURES", mode="overwrite")
