from azure.storage.blob import BlobServiceClient
from io import BytesIO
from azure.storage.blob._blob_client import BlobClient
from typing import List, Optional
import polars as pl


def blob_connector(
    blob_service_client,
    container: str,
    name: str,
) -> BlobClient:
    """
    Returns a blob client for the blob with the given
    name in the given container

    Args:
      blob_service_client: The blob service client instance.
      container (str): The name of the container
        in which the blob is located.
      name (str): The name of the blob.

    Returns:
      A BlobClient object
    """
    container_client = blob_service_client.get_container_client(container)
    return container_client.get_blob_client(name)


def write_blob(
    blob_client: BlobClient,
    source_file: str,
    blob_type: str,
    overwrite: bool,
) -> None:
    """
    Uploads the contents of a file to a blob

    Args:
      blob_client (BlobClient): BlobClient object.
      source_file (str): The path to the file you want to upload.
      blob_type (str): The type of blob to create.
        Either 'BlockBlob', 'PageBlob' or 'AppendBlob'.
      overwrite (bool): If True, overwrite the blob if it already exists.
        If False, fail if the blob already exists.
    """
    with open(source_file, "rb") as data:
        blob_client.upload_blob(
            data,
            blob_type=blob_type,
            overwrite=overwrite,
        )


def download_blob(blob_client: BlobClient, target_file: str):
    """
    Download the blob from the blob client and
    write it to the target file

    Args:
      blob_client (BlobClient): The blob client for the blob
        you want to download.
      target_file (str): The path to the file you want to
        download the blob to.
    """
    with open(target_file, "wb") as file:
        download_stream = blob_client.download_blob()
        file.write(download_stream.readall())


def read_blob(
    blob_client: BlobClient, container: str, name: str
) -> pl.DataFrame:
    """
    Read a blob from a container and return a polars dataframe

     Args:
       blob_client (BlobClient): blob client instance
       container (str): the name of the container in which the blob is stored
       name (str): the name of the blob

     Returns:
       A polars dataframe
    """
    stream = blob_client.download_blob()
    bytes = BytesIO(stream.content_as_bytes())
    return pl.read_parquet(bytes)


class AzureBlob:
    def __init__(self, connection_string: str) -> None:
        self.connection_string = connection_string
        self.blob_service_client = BlobServiceClient.from_connection_string(
            connection_string
        )

    def list(
        self,
        container: str,
    ) -> List[str]:
        container_client = self.blob_service_client.get_container_client(
            container
        )
        return list(container_client.list_blob_names())

    def append(
        self,
        container: str,
        name: str,
        source_file: str,
    ) -> None:
        blob_client = blob_connector(
            blob_service_client=self.blob_service_client,
            container=container,
            name=name,
        )
        write_blob(
            blob_client=blob_client,
            source_file=source_file,
            blob_type="AppendBlob",
            overwrite=False,
        )

    def upload(
        self,
        container: str,
        name: str,
        source_file: str,
        overwrite: bool = False,
    ) -> None:
        blob_client = blob_connector(
            blob_service_client=self.blob_service_client,
            container=container,
            name=name,
        )
        exists = blob_client.exists()
        if not overwrite and exists:
            print("Blob already exists, no permission to overwrite")
        elif (not overwrite and not exists) or overwrite:
            write_blob(
                blob_client=blob_client,
                source_file=source_file,
                blob_type="BlockBlob",
                overwrite=overwrite,
            )

    def download(self, container: str, name: str, target_file: str) -> None:
        blob_client = blob_connector(
            blob_service_client=self.blob_service_client,
            container=container,
            name=name,
        )
        download_blob(blob_client=blob_client, target_file=target_file)

    def read(self, container: str, name: str):
        blob_client = blob_connector(
            blob_service_client=self.blob_service_client,
            container=container,
            name=name,
        )
        return read_blob(
            blob_client=blob_client,
            container=container,
            name=name,
        )
