import holidays
import pandas as pd
import numpy as np
import polars as pl
from typing import Iterable, Dict, Callable, Optional, Union
from datetime import datetime
from dateutil.relativedelta import relativedelta
from dateutil.relativedelta import relativedelta, FR


def blackfriday(
    years: Union[int, Iterable],
) -> Dict:
    """
    Returns a dictionary of Black Friday dates for the given years

    Args:
      years (int | Iterable): a set of years

    Returns:
      A dictionary with a date as the key and a string as the value.
    """
    if isinstance(years, int):
        years = [years]
    return {
        datetime.date(
            datetime(year, 11, 1) + relativedelta(weekday=FR(+4))
        ): "Black Friday"
        for year in years
    }


def get_events(
    years: Union[int, Iterable],
    country: str,
    date_col: str = "DS",
    special: Optional[Dict[str, Callable]] = None,
) -> pd.DataFrame:
    """
    Takes a list of years, a country, and a date column name,
    and returns a dataframe with a date
    column and an event column

    Args:
      years (int | Iterable): a list of years
      country (str): The country for which you want to get the holidays.
      date_col (str): The name of the column that will contain the dates.
        Defaults to "DS"
      special (Optional[Dict[str, Callable]]): An optional dictionary of
        functions to produce special events for the given years

    Returns:
      A dataframe with the date and the event.
    """
    hdict = holidays.country_holidays(country, years=years)
    if special is not None:
        for fun in special.values():
            hdict.update(fun(years=years))
    hdf = (
        pd.DataFrame.from_dict(hdict, orient="index")
        .sort_index()
        .reset_index()
        .rename(columns={"index": date_col, 0: "event"})
    )
    hdf[date_col] = pd.to_datetime(hdf[date_col])
    return hdf


def time_to_events(
    dates: Union[str, pd.Timestamp, Iterable],
    events: Union[pd.DataFrame, pl.DataFrame],
    date_col: str = "DS",
    event_col: str = "event",
) -> pl.DataFrame:
    """
    This function takes a list of dates and a list of events,
    and returns a table with the time to the next event
    and the time since the last event for each date

    Args:
      dates (str | pd.Timestamp | Iterable): a timeline as a date sequence
        events (pd.DataFrame): a dataframe with a date column and an
        event column
      date_col (str): The name of the column in the events
        dataframe that contains the dates of the events. Defaults to "DS"
      event_col (str): The name of the column in the events dataframe
        that contains the event names. Defaults to "event"

    Returns:
      A Polars dataframe with the time to the next event and the time since
      the last event.

    Example:
    >>> import pandas as pd
    >>> dates = pd.date_range("2023-11-20", "2023-12-28")
    >>> events = pd.DataFrame(
        {
            "DS": ["2023-11-23", "2023-11-24", "2023-12-25"],
            "event": ["Thanksgiving", "Black Friday", "Christmas Day"]
        }
    )
    >>> time_to_events(dates=dates, events=events)
    """
    pl.Config.with_columns_kwargs = True
    if isinstance(dates, str):
        dates = pd.to_datetime(dates)
    if not isinstance(events, pl.DataFrame):
        events = pl.DataFrame(events).select(
            [date_col, event_col],
        )  # type: ignore
    if not events[date_col].dtype == pl.Date:
        events = events.with_columns(
            pl.col(date_col)
            .str.strptime(pl.Date, fmt="%Y-%m-%d")
            .alias(date_col)
        )  # type: ignore
    dates = pl.DataFrame({"date": dates})
    if not dates["date"].dtype == pl.Date:
        dates = dates.with_columns(pl.col("date").cast(pl.Date).alias("date"))
    nextd = dates.join_asof(
        events,
        left_on="date",
        right_on=date_col,
        strategy="forward",
    )
    nextd = (
        nextd.with_columns(
            **{"time_to_next_event": pl.col(date_col) - pl.col("date")}
        )
        .select(pl.exclude(date_col))
        .rename({event_col: f"next_{event_col}"})
    )
    prevd = dates.join_asof(
        events,
        left_on="date",
        right_on=date_col,
        strategy="backward",
    )
    prevd = (
        prevd.with_columns(
            **{"time_since_last_event": pl.col("date") - pl.col(date_col)}
        )
        .select(pl.exclude(date_col))
        .rename({event_col: f"last_{event_col}"})
    )
    shared_cols = np.intersect1d(nextd.columns, prevd.columns).tolist()
    diff_cols = (
        np.setdiff1d(nextd.columns, prevd.columns).tolist()
        + np.setdiff1d(prevd.columns, nextd.columns).tolist()
    )

    joined = nextd.join(prevd, on=shared_cols)
    return joined.select(shared_cols + diff_cols).rename({"date": date_col})


class EventFactory:
    def __init__(
        self,
        country: str,
        date_col: str = "DS",
        event_col: str = "event",
        min_date: Optional[str] = None,
        max_date: Optional[str] = None,
        special_events: Optional[Dict[str, Callable]] = None,
    ) -> None:
        """
        EventFactory takes in a country, a date column, an event column, and a
        dictionary of special events.

        Args:
          country (str): The country you want to generate holidays for.
          date_col (str): The name of the column in the dataframe
            that contains the dates. Defaults to "DS"
          event_col (str): The name of the column in the dataframe
            that contains the event names. Defaults to "event"
          min_date (str): optional firts date to use. If not provided,
            ten years before present date will be used.
          max_date (str): optional last date to use. If not provided,
            a year after present date will be used.
          special_events ([Optional[Dict[str, Callable]]]): A dictionary of
            special events. The keys are the names of the events,
            and the values are functions that takes years and return dates
            and the event name as a dictionary.

        Example:
        >>> factory = EventFactory(country="US")
        >>> # Generating a dataframe of events, where each row is a single event, and each column is a
        # single day.
        factory.generate()
        """
        self.country = country
        self.date_col = date_col
        self.event_col = event_col
        if special_events is None:
            self.special = {"Black Friday": blackfriday}
        else:
            self.special = special_events
        if min_date is None:
            min_date = datetime.today() - relativedelta(years=10)
        if max_date is None:
            max_date = datetime.today() + relativedelta(years=1)
        self.min_date = min_date
        self.min_year = min_date.year
        self.max_date = max_date
        self.max_year = max_date.year

    def get_events(
        self,
        years: Optional[Union[int, Iterable]] = None,
        include_observed: bool = False,
        lower_window: int = 1,
        upper_window: int = 0,
        add_special: bool = False,
    ) -> pl.DataFrame:
        """
        Takes a country and a list of years, and returns
        a dataframe with the dates of the holiday/special
        events in those years.

        Args:
          years (int | Iterable): a list of years
          include_observed (bool): Whether to include holidays that are
            "observed" on a different day than the actual holiday.
            Defaults to False
          lower_window (int): A window of influence before event,
            used by Prophet. Defaults to 1
          upper_window (int): A window of influence after event,
            used by Prophet. Defaults to 0
          add_special (bool): If True, add special events,
            defined at class init to holiday events. Defaults to False

        Returns:
          A Polars dataframe with the dates of the events,
          the name of the events, and the lower and upper
          window of influence.
        """
        if years is None:
            years = range(self.min_year, self.max_year + 1)
        if add_special:
            special = self.special
        else:
            special = None
        hdf = get_events(
            years=years,
            country=self.country,
            special=special,
            date_col=self.date_col,
        )
        if not include_observed:
            hdf = hdf.query("event.str.contains('[a-z]$')")
        hdf = hdf.assign(
            **{
                "lower_window": lower_window,
                "upper_window": upper_window,
            }
        )
        hdf = pl.DataFrame(hdf)
        hdf = hdf.with_columns(
            pl.col(self.date_col).cast(pl.Date).alias(self.date_col)
        )
        return hdf

    def time_to_events(
        self,
        events: Union[pd.DataFrame, pl.DataFrame],
        dates: Optional[Union[str, pd.Timestamp, Iterable]] = None,
    ) -> pl.DataFrame:
        """
        Given a list of dates, and returns a table with
        the time to the next event and the time since the
        last event for each date

        Args:
            events (pd.DataFrame | pl.DataFrame): a DataFrame of events.
            dates (str | pd.Timestamp | Iterable): An optinal date or
                list of dates to calculate the time to event. If not
                provided, these will be set automatically from the class
                attributes.

        Returns:
            A Polars DataFrame with the time to event for each date.
        """
        if dates is None:
            dates = pd.date_range(self.min_date, self.max_date, freq="D")
        return time_to_events(
            event_col=self.event_col,
            date_col=self.date_col,
            dates=dates,
            events=events,
        )

    def generate(
        self,
        years: Optional[Union[int, Iterable]] = None,
        include_observed: bool = False,
        lower_window: int = 1,
        upper_window: int = 0,
        add_special: bool = False,
    ) -> pl.DataFrame:
        """
        This function returns a dataframe of events, where each row is a
        single event, and each column is a single day

        Args:
          years (Optional[Union[int, Iterable]]): The years to generate
            events for. If None, these will be taken from class attributes.
          include_observed (bool): Whether to include the observed events
            in the generated data. Defaults to False
          lower_window (int): the length of effective window defore event,
            used by Prophet. Defaults to 1
          upper_window (int): the length of effective window after event,
            used by Prophet. Defaults to 0
          add_special (bool): If True, adds special events to the dataframe.
            Defaults to False

        Returns:
          A polars dataframe with holiday features
        """
        events = self.get_events(
            years=years,
            include_observed=include_observed,
            lower_window=lower_window,
            upper_window=upper_window,
            add_special=add_special,
        )
        return self.time_to_events(events=events)
