import snowflake.connector
from snowflake.connector.connection import SnowflakeConnection
import yaml
import pandas as pd
import polars as pl
from typing import Union


# Utility functions:
def list_tables(connection: SnowflakeConnection) -> pd.DataFrame:
    """
    This function takes a Snowflake connection and returns
    a pandas dataframe of all the tables in the
    database and schema of the connection

    Args:
      connection (SnowflakeConnection): Snowflake connection instance

    Returns:
      A dataframe with the database, schema, and table name.
    """
    connection.cursor().execute("show tables;")
    tabs = connection.cursor().execute(
        """
        select "name" from table(result_scan(last_query_id()));
        """
    )
    tables = tabs.fetch_pandas_all()
    tables.insert(0, "database", connection.database)
    tables.insert(1, "schema", connection.schema)
    return tables


def run_query(
    connection: SnowflakeConnection,
    query_string: str,
    as_pandas: bool = False,
) -> Union[pd.DataFrame, pl.DataFrame]:
    """
    This function takes a query string and a connection object,
    and returns a either a polars or pandas DataFrame.

    Args:
      connection (SnowflakeConnection): Snowflake connection instance
      query_string (str): The SQL query to run
      as_pandas (bool): If True, the output will be a pandas DataFrame.
        If False, the output will be a polars DataFrame. Defaults to False

    Returns:
      A DataFrame object
    """
    res = connection.cursor().execute(query_string)
    tab = res.fetch_arrow_all()
    if as_pandas:
        out = tab.to_pandas()
    else:
        out = pl.DataFrame(tab)
    return out
